/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

control msgs:
(output-to-file <file>)
(append-output-to-file <file>)
(end-file-output)
(file0 <file>)
(file1 <file>)
(rate0 <rate>)
(rate1 <rate>)
(position0 <position>)
(position1 <position>)
(position0-get)
(position1-get)
(position0-delta <delta>)
(position1-delta <delta>)
(setloop0 <start> <end>)
(clearloop0)
(setloop1 <start> <end>)
(clearloop1)
(cue0)
(cue1)
(level0 <rate>)
(level1 <rate>)
(status)
(shutdown)

*/

#include <cstdlib>
#include <exception>
#include <cstddef>
#include <new>
  extern "C" {
#include <unistd.h>
#include <stdio.h>
#include <sched.h>
#include <math.h>
#include <sys/mman.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <pthread.h>
	     }
#include <string>
#include <iostream>

using namespace std;

#define ASYNCH
#ifdef ASYNCH
#include "wavfilea.h"
#else
#include "wavfile.h"
#endif

#include "soundout.h"
#include "sexpr.h"
#include "char_ready.h"
#include "dprintf.h"


const char progname[] = "mixd";

class EndApp : public exception {};

void
mix(size_t x, float *fbuf, float *fbuf1, float *fbufm, 
    float level, float level1)
{
  for (size_t i = 0; i < x*2; i++)
    *fbufm++ = *fbuf++ * level + *fbuf1++ * level1;
}

float rate0 = 1.0, rate1 = 1.0, level0, level1;

// based on two players..this is the max of the output you'd want each to be
float level_share = .5;

bool cue0 = false, cue1 = false;
bool cuep = false;

void
spew_status()
{
}

#ifdef ASYNCH
wavfilea *player0, *player1;
#else
soundfile *player0, *player1;
#endif

FILE *outfile;

// make this table driven?
void
handle_command(sexpr s)
{
  if (s.pairp())
    {
      sexpr a(s.car());
      sexpr l(s.cdr());
      
      if (sexpr::eq(a, sexpr::symbol("shutdown")))
	{
	  throw EndApp();
	}
      if (sexpr::eq(a, sexpr::symbol("end-file-output")))
	{
	  if (outfile)
	    {
	      fclose(outfile);
	      outfile = 0;
	      printf("(end-file-output)\n");
	      fflush(stdout);
	    }
	  else
	    {
	      printf("(error \"no current output file\")\n");
	      fflush(stdout);
	    }
	}
      if (sexpr::eq(a, sexpr::symbol("output-to-file")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (outfile)
		    fclose(outfile);
		  if (!(outfile = fopen(string(f).c_str(), "wb")))
		    {
		      printf("(error \"couldn't open %s for writing: %s\")\n",
			     string(f).c_str(), strerror(errno));
		      fflush(stdout);
		    }
		  else
		    {
		      printf("(output-to-file \"%s\")\n", string(f).c_str());
		      fflush(stdout);
		    }
		}
	    }
	}
      if (sexpr::eq(a, sexpr::symbol("append-output-to-file")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (outfile)
		    fclose(outfile);
		  if (!(outfile = fopen(string(f).c_str(), "ab")))
		    {
		      printf("(error \"couldn't open %s for appending: %s\")\n",
			     string(f).c_str(), strerror(errno));
		      fflush(stdout);
		    }
		  else
		    {
		      printf("(append-output-to-file \"%s\")\n", 
			     string(f).c_str());
		      fflush(stdout);
		    }
		}
	    }
	}
      if (sexpr::eq(a, sexpr::symbol("file0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (player0)
		    delete player0;
		  try
		    {
#ifdef ASYNCH
		      player0 = new wavfilea(f, 44100, 2);
#else
		      player0 = new wavfile(f, 44100, 2);
#endif
                      printf("(file0 \"%s\" %f)\n", 
			     string(f).c_str(), player0->size());
                      fflush(stdout);
		    }
		  catch (...)
		    {
		      cerr << "problem loading player0" << endl;
		      player0 = 0;
		      printf("(error \"couldn't load %s on player0\")\n",
			     string(f).c_str());
                      fflush(stdout);
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("file1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (player1)
		    delete player1;
		  try
		    {
#ifdef ASYNCH
		      player1 = new wavfilea(f, 44100, 2);
#else
		      player1 = new wavfile(f, 44100, 2);
#endif
                      printf("(file1 \"%s\" %f)\n", 
			     string(f).c_str(), player1->size());
                      fflush(stdout);
		    }
		  catch (...)
		    {
		      cerr << "problem loading player1" << endl;
		      player1 = 0;
		      printf("(error \"couldn't load %s on player1\")\n",
			     string(f).c_str());
                      fflush(stdout);
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("cue0")))
	{
	  cue0 = !cue0;
	}
      else if (sexpr::eq(a, sexpr::symbol("cue1")))
	{
	  cue1 = !cue1;
	}
      else if (sexpr::eq(a, sexpr::symbol("rate0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		rate0 = f;
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("rate1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		rate1 = f;
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("setloop0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  sexpr l1(l.cdr());
		  if (l1.pairp() && l1.car().floatp())
		    {
		      if (player0)
			player0->setloop(f, l1.car());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("setloop1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  sexpr l1(l.cdr());
		  if (l1.pairp() && l1.car().floatp())
		    {
		      if (player1)
			player1->setloop(f, l1.car());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("clearloop0")))
	{
	  if (player0)
	    player0->clearloop();
	}
      else if (sexpr::eq(a, sexpr::symbol("clearloop1")))
	{
	  if (player1)
	    player1->clearloop();
	}
      else if (sexpr::eq(a, sexpr::symbol("position0-delta")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player0)
		  {
		    float f1 = player0->tell();
		    float f2 = float(f) + f1;
		    dprintf(1, "setting player0 to %f = %f + %f\n",
			    f2, f1, float(f));
		    player0->seek(f2);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position1-delta")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player1)
		  {
		    float f1 = player1->tell();
		    dprintf(1, "setting player1 to %f = %f + %f\n",
			    f1 + float(f), f1, float(f));
		    player1->seek(float(f) + f1);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position0-get")))
	{
	  fprintf(stdout, "(position0 %f)\n", (player0 ? player0->tell() : 0.0));
	  fflush(stdout);
	}
      else if (sexpr::eq(a, sexpr::symbol("position1-get")))
	{
	  fprintf(stdout, "(position1 %f)\n", (player1 ? player1->tell() : 0.0));
	  fflush(stdout);
	}
      else if (sexpr::eq(a, sexpr::symbol("position0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player0)
		  {
		    dprintf(1, "setting player0 to %f\n", float(f));
		    player0->seek(f);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player1)
		  player1->seek(f);
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("level0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  float f1 = f;
		  if (f1 > 1)
		    f1 = 1;
		  if (f1 < 0)
		    f1 = 0;
		  level0 = f1 * level_share;
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("level1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  float f1 = f;
		  if (f1 > 1)
		    f1 = 1;
		  if (f1 < 0)
		    f1 = 0;
		  level1 = f1 * level_share;
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("status")))
	spew_status();
    }
}

void
doio()
{
  while (char_ready(0))
    {
      sexpr s;

      cin >> s;

      /* NB: this ensures we don't go into a tight loop of a
	non-blocking select() followed by a non-blocking read..this is
	critical for a real-time program!  */
      if (s.eof_objectp())
	throw EndApp();

      handle_command(s);
    }
}

const char *cue_channel = 0;

bool pci128_lossage = false;
int fragment = 0x00070009;

void
pci128fix(float *f, size_t n)
{
  for (size_t i = 0; i < n; i++)
    {
      f[2*i+1] = (f[2*i] + f[2*i+1])/2;
      f[2*i] = 0;
    }
}

struct option options[] =
{{"fragment", 1, 0, 'f'},
 {"pci128-lossage", 0, 0, 'p'},
 {"cue-channel", 1, 0, 'c'},
 {"debug", 1, 0, 'd'},
 {0, 0, 0, 0}};

#ifdef __linux__
extern "C" int fileno(FILE *);
extern "C" int sync();
extern "C" void usleep(unsigned long);
#endif

extern "C"
void *
syncer(void *)
{
  while (true)
    {
      // !!!! if the syncing helps we'll need a mutex for outfile
      if (outfile)
	sync(fileno(outfile));
      usleep(1000000);
    }
}
  
// mixd
int
main1(int argc, char **argv)
{
  int idx;

  int rc;

  bool clerror = false;

  while ((rc = getopt_long(argc, argv, "c:f:p", options, &idx)) != EOF)
    switch (rc)
      {
      case ':': 
      case '?': clerror = true; break;
      case 'c': 
	cue_channel = optarg; 
	break;
      case 'p': 
	pci128_lossage = true; 
	break;
      case 'f': 
	{
	  int i;
	  char c;
	  
	  if (sscanf(optarg, "%x%c", &i, &c) == 1)
	    {
	      i &= 0xFFFFFFFF;
	      fragment = i;
	    }
	  else
	    {
	      dprintf(0, "%s: malformed argument to --fragment: %s\n",
		      progname, optarg);
	      clerror = true;
	    }
	break;
	}
      case 'd': 
	{
	  int i;
	  char c;
	  
	  if (sscanf(optarg, "%d%c", &i, &c) == 1)
	    {
	      debuglevel = i;
	      dprintf(0, "debuglevel = %d\n", debuglevel);
	    }
	  else
	    {
	      dprintf(0, "%s: malformed argument to --debug: %s\n",
		      progname, optarg);
	      clerror = true;
	    }
	break;
	}
      }

  if (clerror)
    return 1;

  dprintf(1, "cue channel: %s\n", (cue_channel ? cue_channel : "<none>"));
  dprintf(1, "pci128-lossage = %s\n", (pci128_lossage ? "yes" : "no"));
  dprintf(1, "fragment = 0x%8.8lx\n", fragment);

  if (geteuid() == 0)
    {
      struct sched_param sp;
      sp.sched_priority = 50;
      if (sched_setscheduler(0, SCHED_RR, &sp) == -1)
	{
	  dprintf(1, "couldn't set real-time scheduling: %s\n",
		  strerror(errno));
	  return 1;
	}

#ifdef _POSIX_MEMLOCK
      if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1)
	dprintf(1, "problem locking memory: %s\n", strerror(errno));
#endif
    }

  // sigh..this doesn't help..even with a latency of 200 ms!
  // i tried it with sync freqs of 1, 5, 10 and 20 times a second
#if 0
  // create a syncer thread to attempt to ensure that the physical
  // writes of outfile aren't unduly clustered..if this helps, we'll
  // need a mutex around outfile
  pthread_t synct;
  pthread_attr_t att;
  pthread_attr_init(&att);
  // make sure the thread inherits our scheduling
  pthread_attr_setinheritsched(&att, PTHREAD_INHERIT_SCHED);

  if (pthread_create(&synct, &att, syncer, 0) == -1)
    {
      fprintf(stderr, "couldn't create syncer: %s\n", strerror(errno));
      return 1;
    }

  pthread_attr_destroy(&att);
#endif

  soundout *o = mydsp(fragment);
  if (cue_channel && (cuep = cue_try(cue_channel, fragment)))
    dprintf(1, "cue channel on\n");

  float *fbuf0 = new float [o->getbufsiz()/2];
  float *fbuf1 = new float [o->getbufsiz()/2];
  float *fbufm = new float [o->getbufsiz()/2];
  float *fbufz = new float [o->getbufsiz()/2];

  for (size_t i = 0; i < o->getbufsiz()/2; i++)
    fbufz[i] = 0;

  level0 = level_share;
  level1 = level_share;

  size_t n = o->getbufsiz()/4;

#if 0
#ifdef ASYNCH
  player0 = new wavfilea("ms.wav", 44100, 2);
#else
  player0 = new wavfile("ms.wav", 44100, 2);
#endif
#endif

  while (true)
    {
#ifdef ASYNCH
      bool p0 = player0 && player0->goodtogetp();
      bool p1 = player1 && player1->goodtogetp();
#else
      bool p0 = player0;
      bool p1 = player1;
#endif

      if (p0)
	player0->get(fbuf0, n, rate0);
      if (p1)
	player1->get(fbuf1, n, rate1);

      mix(n, 
	  (p0 ? fbuf0 : fbufz), 
	  (p1 ? fbuf1 : fbufz), 
	  fbufm, 
	  level0, level1);
      o->put(fbufm, n);

      if (outfile)
	outfile_put(outfile, fbufm, n);

      if (cuep)
	{
	  mix(n, 
	      (p0 ? fbuf0 : fbufz), 
	      (p1 ? fbuf1 : fbufz), 
	      fbufm, 
	      (cue0 ? .5 : 0),
	      (cue1 ? .5 : 0));

	  if (pci128_lossage)
	    pci128fix(fbufm, n);

	  cue_put(fbufm, n);
	}

      doio();
    }
  
  return 0;
}
  
int
main(int argc, char **argv)
{
  fflush(stdout);
  try
    {
      return main1(argc, argv);
    }
  catch (bad_alloc)
    {
      cerr << "Out of memory!" << endl;
      return 1;
    }
  catch (EndApp)
    {
      return 0;
    }
  catch (exception &e)
    {
      cerr << "Unhandled exception: " << e.what() << endl;
      return 1;
    }
  catch (...)
    {
      cerr << "Unknown unhandled exception!" << endl;
      return 1;
    }
      
  return 0;
}
