#!/usr/bin/python
#
#    oolaboola -- user interface for digital dj program
#    Copyright (C) 1999  Eric S. Tiedemann
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# TODO:
#  make interface invariant if we make sexprs a builtin type

import sys
import string
import sexprlow

class Symbol:
 tab = {}

 def __init__(self, name):
  self.name = name

 def __str__(self):
  return self.name

def string2symbol(name):
  try:
    return Symbol.tab[name]
  except KeyError:
    Symbol.tab[name] = Symbol(name)
    return Symbol.tab[name]

class String:

 def __init__(self, s):
  self.s = s

 def __str__(self):
  return self.s

class Boolean:
  def __init__(self, truep):
    self.truep = truep

  def __str__(self):
    if (self.truep):
      return "#t"
    else:
      return "#f"

True = Boolean("true")
False = Boolean(None)  

class pair:
  def __init__(self, a, b):
    self.a = a
    self.b = b
  def car(self):
    return self.a
  def cdr(self):
    return self.b
  def prtail(self, b):
    if (isinstance(b, pair)):
      return (" %s" % str(b.car())) + self.prtail(b.cdr())
    elif (b == None):
      return ")"
    else:
      return " . %s)" % str(b)

  def __str__(self):
    if (self.b == None):
      return "(%s)" % str(self.a)
    else:
      return ("(%s" % str(self.a)) + self.prtail(self.b)

eofobject = ["eofobject"]

def eofobjectp(o):
  return o == eofobject

closep = ["closep"]
dot = ["dot"]
openp = ["openp"]
quote = ["quote"]

def read_list(f):
  o = read1(f)
  if (o == eofobject):
    raise EOFError
  
  if (o == closep):
    return None
  elif (o == dot):
    # hmm..but we really don't want a dot at the start of a list
    o = read(f)
    if (o == eof_object):
      raise EOFError
    e = read1(f)
    if (e == closep):
      return o
    elif (e == eofobject):
      raise EOFError
    else:
      raise "SyntaxError", "Malformed dotted list syntax."
  else:
    b = read_list(f)
    return pair(o, b)
  
def read1(f):
  (o, t) = sexprlow.rdatom(f)

  if (t == 0):
    return o
  elif (t == 1):
    return string2symbol(o)
  elif (t == 2):
    return eofobject
  elif (t == 3):
    return dot
  elif (t == 4):
    return read_list(f)
  elif (t == 5):
    return closep
  elif (t == 6):
    return String(o)
  elif (t == 7):
    raise EOFError, "end-of-file in the middle of a string"

def read(f):
  o = read1(f)
  
  if (o == dot):
    raise "SyntaxError", "Superfluous period found in input."
  elif (o == closep):
    raise "SyntaxError", "Superfluous close paren found in input."
  else:
    return o

def escape_string(s):
    s1 = "\""
    for c in s:
        if (c == "\"" or c == "\\"):
            s1 = s1 + "\\" + c
        else:
            s1 = s1 + c
    s1 = s1 + "\""
    return s1

if (__name__ == '__main__'):
  e = read(sys.stdin)
  while (not e == eofobject):
    #print e
    e = read(sys.stdin)

def foo(fl):
  f = open(fl)
  e = read(f)
  while (not e == eofobject):
    e = read(f)
  f.close()
