/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <python1.5/Python.h>
#include <ctype.h>
#include <stdlib.h>

/*

  sexprlow.rdatom(fileobj) -> (obj, integer)
  the integer takes on the following values:
  0 if obj is a float
  1 if obj is a string representing a symbol
  2 if eof was encountered
  3 if a dot was read
  4 if an open paren was read
  5 if a close paren was read
  6 if a string was read (obj is the string)
  7 if an unterminated string was read

  TODO: acquire a ref to the file object?
        read string constants

*/

#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE
#endif

static PyObject *
rdatom(FILE *fp)
{
  int c;
  double d;
  char *end;
  char buf[1024];
  int bufi;
  int instring = 0;

  bufi = 0;

  while (bufi < sizeof(buf) - 1)
    {
      Py_BEGIN_ALLOW_THREADS
      c = getc(fp);
      Py_END_ALLOW_THREADS
      switch (c)
	{
	case EOF: goto looped; break;
	case '(':
	  if (bufi == 0)
	    return Py_BuildValue("si", "", 4);
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	  break;
	case ')':
	  if (bufi == 0)
	    return Py_BuildValue("si", "", 5);
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	  break;
	case '\"':
          if (!instring && bufi == 0)
	    {
	      instring = 1;
	    }
	  else if (instring)
	    {
	      goto looped;
	    }
          else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	  break;
	case ';':
	  if (bufi == 0)
	    {
	      Py_BEGIN_ALLOW_THREADS
	      while ((c = getc(fp)) != EOF && c != '\n')
		;			/* consume comment */
	      Py_END_ALLOW_THREADS
	    }
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	  break;
	default:
	  if (isspace(c))
	    {
	      if (bufi > 0)
		goto looped;
	    }
	  else
	    buf[bufi++] = c;
	}
    }

 looped:
  buf[bufi] = 0;

  if (instring)
    {
      if (c == EOF)
	;			/* complain */
      return Py_BuildValue("si", buf, 6);
    }

  if (bufi == 0)
    return Py_BuildValue("si", buf, 2);

  if (bufi == 1 && *buf == '.')
    return Py_BuildValue("si", buf, 3);
    
  d = strtod(buf, &end);

  if (end == buf + strlen(buf))
    return Py_BuildValue("fi", d, 0);
  else
    return Py_BuildValue("si", buf, 1);
}

static PyObject *
sexprlow_rdatom(PyObject *self, PyObject *args)
{
  PyObject *f;

  if (!PyArg_ParseTuple(args, "O!;sexprlow.rdatom", &PyFile_Type, &f))
    return NULL;

  return rdatom(PyFile_AsFile(f));
}

static PyMethodDef SexprlowMethods[] =
{
  {"rdatom", sexprlow_rdatom, METH_VARARGS},
  {0, 0, 0}
};

void
initsexprlow()
{
  Py_InitModule("sexprlow", SexprlowMethods);
}
